/**
 * Coordinate.java:	encapsulates a vector (or coordinate)
 *
 * @author Sven Sa 
 * @version V1.0 06/06/97
 */
public class Coordinate
{
	public final static int NORMAL_BASE = 1024;	// the length of a normal Vector
	public final static Coordinate VIEW_DIR = new Coordinate (0,0,NORMAL_BASE); // normal view direction

	public int x,y,z;  // the coordinates

	/**
	 * constructs an empty coordinate
	 *
	 * @param void
	 */
	public Coordinate()
	{
		x=y=z=0;
	}

	/**
	 * constructs a coordinate
	 *
	 * @param x_ x-axis
	 * @param y_ y-axis
	 * @param z_ z-axis
	 */
	public Coordinate(int x_, int y_, int z_)
	{
		x=x_;
		y=y_;
		z=z_;
	}

	/**
	 * constructs a coordinate (copy constructor)
	 *
	 * @param coord xyz-axis
	 */
	public Coordinate(Coordinate coord)
	{
		x=coord.x;
		y=coord.y;
		z=coord.z;
	}

	/**
	 * calculates the cross product of two vectors
	 *
	 * @param c1 first vector
	 * @param c2 second vector
	 * @return Coordinate the cross product
	 */
	public final static Coordinate cross (Coordinate c1,Coordinate c2)
	{
		return new Coordinate (c1.y * c2.z - c1.z * c2.y,
			 				   c1.z * c2.x - c1.x * c2.z,
							   c1.x * c2.y - c1.y * c2.x);
	}															  

/*	public final static int zcrossProduct (Coordinate c1,Coordinate c2)
	// calculates the z part of cross product from two vectors
	{
		return c1.x * c2.y - c1.y * c2.x;
	}*/

	/**
	 * calculates the dot product of two vectors
	 *
	 * @param c1 first vector
	 * @param c2 second vector
	 * @return Coordinate the dot product
	 */
	public final static int dot (Coordinate c1,Coordinate c2)
	{
		return (c1.x * c2.x + c1.y * c2.y + c1.z * c2.z);
	}

	/**
	 * scales a vector by given factor
	 *
	 * @param c the vector
	 * @param factor the factor to scale by
	 * @return Coordinate the scaled vector
	 */
	public final static Coordinate scale (Coordinate c,int factor)
	{
		return new Coordinate (c.x *= factor,
							   c.y *= factor,
							   c.z *= factor);
	}

	/**
	 * normalizes a vector to length of NORMAL_BASE
	 *
	 * @param cor the vector
	 * @return Coordinate the normalized vector
	 * @see NORMAL_BASE
	 */
	public final static Coordinate normalize (Coordinate cor)
	{
		Coordinate	ret = new Coordinate (cor);		
		if (length(cor) < 100) ret =scale (ret,100);// if length too small possible rounding errors!
		int			len = length(ret);
		
		if( len != 0 )	// if it's a point (nullvector) you cannot tell it's direction
		{
			ret.x = ret.x * NORMAL_BASE / len;
			ret.y = ret.y * NORMAL_BASE / len;
			ret.z = ret.z * NORMAL_BASE / len;
		}
		return ret;
	}

	/**
	 * returns the length of a vector
	 *
	 * @param cor the vector
	 * @return int the length
	 */
	public final static int length (Coordinate cor)
	// returns the length of a vector
	{
	    return (int)(Math.sqrt( dot(cor,cor) ));
	}

	/**
	 * subtracts a vector from another vector
	 *
	 * @param c1 the first vector
	 * @param c2 the second vector
	 * @return Coordinate c1-c2
	 */
	public final static Coordinate sub (Coordinate c1, Coordinate c2)
	{
		return new Coordinate (c1.x - c2.x, 
							   c1.y - c2.y, 
							   c1.z - c2.z);
	}

	/**
	 * adds a vector to another vector
	 *
	 * @param c1 the first vector
	 * @param c2 the second vector
	 * @return Coordinate c1-c2
	 */
	public final static Coordinate add (Coordinate c1, Coordinate c2)
	{
		return new Coordinate (c1.x + c2.x, 
							   c1.y + c2.y, 
							   c1.z + c2.z);
	}
}
