//******************************************************************************
// FastRotation.java:	Eine mgliche Implementation des RotationEngine Interfaces
//
// Author:  Sven Sa 
// am:		20/05/97
// Besch.:	Implementiert RotationEngine. Hat die Aufgabe Punkte zu rotieren.
//			Um eine aktzeptable Performance zu erreichen wird dabei auf eine
//			Cosinus-/Sinustabelle zurckgegriffen. Es wird angenommen, da es nicht
//			notwendig ist absolut przise zu rechnen (8 Bit Krzung).
//			Die Cosinus-/Sinustabelle ist um PREC nach links geshiftet, um zu 
//			vermeiden mit Gleitkommazahlen zu arbeiten. (Je nach Anwendung ist
//			darauf zu achten, ob die Vorraussetzung angemessen erscheinen. (Bei
//			groen Vektorobjekten kann es zu (nicht abgefangenen) berlufen 
//			kommen.
// 
//******************************************************************************

import Coordinate;			// die Koordinatenklasse

public class FastRotate implements RotationEngine
{
	//-- Variablen -----------------------------------------------------------------
	public static final double pi = 3.14159265;		// ?!?

	private Coordinate Rotation;				// die aktuelle Rotation
	private int CosX,CosY,CosZ,SinX,SinY,SinZ;	// die Cos/Sin Werte der aktuellen Rot.

	private static final int PREC = 8;				// um 8 Bits in den Tabellen shiften
	private static final int TABLE_SIZE = 360 * 4;	// 1/4 Grad Rotation mglich
	private static final int TABLE_SIZE_eff = TABLE_SIZE + (TABLE_SIZE/4); // effektive Gre
	private static int [] CS_Table;						// Die Cosinus-/Sinustabelle

	//-- Konstruktor ---------------------------------------------------------------

	public FastRotate ()
	// die Rotation auf Null setzen
	{
		Rotation = new Coordinate(); 
		initCS();
	}

	public FastRotate (Coordinate coord)
	// die Rotation auf gegebene Koordinaten setzen
	{
		Rotation = new Coordinate(coord.x,coord.y,coord.z); 
		initCS();
	}

	public FastRotate (int x, int y, int z)
	// die Rotation auf gegebene Koordinaten setzen
	{
		Rotation = new Coordinate(x,y,z); 
		initCS();
	}

	//-- (public) Methoden ----------------------------------------------------------

	private void initCS()
	// Initialisierte die Cosinus-/Sinustabelle (fllt sie mit Sinuswerten von
	// 0 bis 359+90 Grad in 1/4 Grad Schritten
	{
		CS_Table = new int[TABLE_SIZE_eff]; // Speicher reservieren

		try
		{
			for (int i=0; i<TABLE_SIZE_eff; ++i)
				CS_Table[i]= (int)(Math.sin ((double)i/2/360 * pi)*(1<<PREC)); // 1/4 Grad Schritte
		}
		catch (IndexOutOfBoundsException e)
		{
			System.err.println ("Schwerwiegender Fehler: FastRotate.initCS: IndexOutOfBounds!");
		}
		catch (Exception e)
		{
			System.err.println ("Schwerwiegender Fehler: FastRotate.initCS: " + e.toString());
		}

	}

	//-- (public) Methoden ----------------------------------------------------------

	/**
	 * calculate:		Rotiert die Koordinate
	 *
	 *	PARAMETER:
	 *		x_rot,
	 *		y_rot,
	 *		z_rot	-	Abstand auf der X,Y,Z-Achse
	 *
	 * 	RETURNS:
	 *		coord	-	die rotierten Koordinaten zusammengefat zu einer Koordinate
	 **/
	public Coordinate calculate (int x_rot,int y_rot, int z_rot )
	{
		Coordinate coord = new Coordinate();
		coord.y=(y_rot*CosX - z_rot*SinX)>>PREC;
  	    coord.z=(y_rot*SinX + z_rot*CosX)>>PREC;

		coord.x=(x_rot*CosY - coord.z*SinY)>>PREC;
		coord.z=(x_rot*SinY + coord.z*CosY)>>PREC;

		coord.x=(coord.x*CosZ - coord.y*SinZ)>>PREC;
		coord.y=(coord.x*SinZ + coord.y*CosZ)>>PREC;
		return coord;
	}


	/**
	 * calculate: otiert die Koordinate 
	 *
	 *	PARAMETER:
	 *		rot		-	Abstand auf der X,Y,Z-Achse
	 *
	 * 	RETURNS:
	 *		coord	-	die rotierten Koordinaten zusammengefat zu einer Koordinate
	 **/
	public Coordinate calculate (Coordinate rot)
	{
		return calculate (rot.x,rot.y,rot.z);
	};


	/**
	 * calculate:		Rotiert die Koordinate 
	 * 	 
	 *	PARAMETER:
	 *		rot[]	-	Abstnde auf der X,Y,Z-Achse
	 *		cnt		-	Anzahl Elemente in dem Array
	 *
	 * 	RETURNS:
	 *		coord	-	die rotierten Koordinaten zusammengefat zu einer Koordinate
	 **/
	public void calculate (Coordinate []src,Coordinate []dst,int cnt)
	{
		for (int i=0; i<cnt; ++i)					// ber alle Koordinaten
		{
			try
			{
				dst[i]=calculate(src[i]);				// Koordinate rotieren
			}
			catch (IndexOutOfBoundsException e)		// wenn cnt falsch angegeben
			{
				System.err.println ("FastRotate.Rotate(src,dst,cnt) Out of Bounds: " + e.toString());
			}
			catch (Exception e)					   // alle hheren Exceptions fangen
			{
				System.err.println ("FastRotate.Rotate(src,dst,cnt) Exception: " + e.toString());
			}
		}
	};

	public void rotate (int x_rot,int y_rot, int z_rot )
	// sets the relative rotation
	{
		setRotation (Rotation.x + x_rot,
					 Rotation.y + y_rot,
					 Rotation.z + z_rot);
	}

	public void rotate (Coordinate rot )
	// sets the relative rotation
	{
		rotate (rot.x,rot.y,rot.z);
	}

	/**
	 * setRotation:		setzt die Rotation der Klasse (alle folgenden Rotationen bis
	 *					zum nchsten setRotation beziehen sich auf diese Rotation
	 *
	 *	PARAMETER:
	 *		x_rot,
	 *		y_rot,
	 *		z_rot	-	Rotationswinkel der X,Y,Z-Achse (absolut)
	 *
	 * 	RETURNS:
	 *		void
	 **/
	public void setRotation (int x_rot,int y_rot, int z_rot )
	{
		Rotation.x = x_rot; // keine neue generieren, da Methode oft aufgerufen
		Rotation.y = y_rot;	// wird und sonst zuviel Garbage Collection durch-
		Rotation.z = z_rot;	// gefhrt werden mu

		try
		{
			// Anpassung der Werte an die CS_Table 
			if (x_rot<0) x_rot = (TABLE_SIZE - (-x_rot % TABLE_SIZE)); // wenn kleine
			if (y_rot<0) y_rot = (TABLE_SIZE - (-y_rot % TABLE_SIZE));
			if (z_rot<0) z_rot = (TABLE_SIZE - (-z_rot % TABLE_SIZE));

			if (x_rot>=TABLE_SIZE) x_rot = (x_rot % TABLE_SIZE);	   // wenn grer
			if (y_rot>=TABLE_SIZE) y_rot = (y_rot % TABLE_SIZE);
			if (z_rot>=TABLE_SIZE) z_rot = (z_rot % TABLE_SIZE);

			CosX = CS_Table [x_rot+90*4]; // Cosinus ist Phasenverschoben:
			SinX = CS_Table [x_rot];	  // cos(a) = sin(a+90)

			CosY = CS_Table [y_rot+90*4];
			SinY = CS_Table [y_rot];

			CosZ = CS_Table [z_rot+90*4];
			SinZ = CS_Table [z_rot];
			
		} 
		catch (IndexOutOfBoundsException e)
		{
			System.err.println ("FastRotate.setRotation(x,y,z) Out of Bounds: "+ e.toString());
		}
	};


	/**
	 * setRotation:		setzt die Rotation der Klasse (alle folgenden Rotationen bis
	 *					zum nchsten setRotation beziehen sich auf diese Rotation
	 *
	 *	PARAMETER:
	 *		rot			-	Rotationskoordinaten der X,Y,Z-Achse (absolut)
	 *
	 * 	RETURNS:
	 *		void
	 **/
	public void setRotation (Coordinate rot)
	{
		setRotation (rot.x,rot.y,rot.z);
	};


	/**
	 * getRotation:		liest die aktuelle Rotationseinstellung
	 *
	 *	PARAMETER:
	 *		keine
	 *
	 * 	RETURNS:
	 *		rot		- Rotationskoordinaten der Klasse
	 **/
	public Coordinate getRotation ( )
	{
		return new Coordinate(Rotation);
	};


}
